@@ -47,6 +47,10 @@ class MockExtensionApp(ExtensionAppJinjaMixin, ExtensionApp):
     mock_trait = Unicode('mock trait', config=True)
     loaded = False
 
+    @staticmethod
+    def get_extension_package():
+        return "jupyter_server.tests.extension.mockextensions"
+
     def initialize_handlers(self):
         self.handlers.append(('/mock', MockExtensionHandler))
         self.handlers.append(('/mock_template', MockExtensionTemplateHandler))
